﻿using System;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;
using System.Security.Cryptography;
using System.Linq;
namespace ConsoleApp4
{
    public class Program
    {
        private static string userCdrivePath = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.UserProfile), @"C:\\project2creds\");
        private const string plainTextUserCreds = "user_credentials_pt.txt";
        private const string hashedUserCreds = "user_credentials_hash.txt";
        private const string hashSaltUserCreds = "user_credentials_hash_salt.txt";
        private const int maxPasswordLength = 15;
        private const int minPasswordLength = 1;
        private const string passwordFlag = "password";
        private const string userNameFlag = "username";
        private static string deliveryPathPlainTextFile = Path.Combine(userCdrivePath, plainTextUserCreds);
        static string deliveryPathHashFile = Path.Combine(userCdrivePath, hashedUserCreds);
        static string deliveryPathHashSalt = Path.Combine(userCdrivePath, hashSaltUserCreds);

        static string[] retrievePasswordAndSalt(string filePath)
        {
            string[] lines = File.ReadAllLines(filePath);
            foreach (string line in lines)
            {
                string[] parts = line.Split(',');
                return parts;
            }
            return new string[0];
        }
        static bool ReadPassword(string userName,string enteredPassword,string filePath)
        {
            string[] lines = File.ReadAllLines(filePath);
            if(filePath == deliveryPathPlainTextFile)
            {
                foreach (string line in lines)
                {
                    string[] parts = line.Split(',');
                    if (parts.Length == 2)
                    {
                        var storedUsername = parts[0];
                        var storedPassword = parts[1];
                        if (String.Equals(userName, storedUsername, StringComparison.OrdinalIgnoreCase))
                        {
                            var passWordCheck = String.Equals(enteredPassword, storedPassword);
                            if (passWordCheck)
                            {
                                return true;
                            }
                        }
                    }
                    else
                    {
                        Console.WriteLine("Plaintext Password check failed");
                        return false;
                    }
                }
            }else if(filePath == deliveryPathHashFile)
            {
                foreach (string line in lines)
                {
                    string[] parts = line.Split(',');
                    if (parts.Length == 2)
                    {
                        var storedUsername = parts[0];
                        var storedHashPassword = parts[1];
                        if (String.Equals(userName, storedUsername, StringComparison.OrdinalIgnoreCase))
                        {
                            var hashEnteredPassword = ComputeHash(enteredPassword);
                            var passWordCheck = String.Equals(hashEnteredPassword, storedHashPassword);
                            if (passWordCheck)
                            {
                                return true;
                            }
                        }
                    }
                    else
                    {
                        Console.WriteLine("Hashed Password check failed");
                        return false;
                    }
                }
            }
            else if(filePath == deliveryPathHashSalt)
            {
                foreach (string line in lines)
                {
                    string[] parts = line.Split(',');
                    if (parts.Length == 3)
                    {
                        var storedUsername = parts[0];
                        var storedHashPassword = parts[1];
                        var storedSalt = parts[2];
                        var joinedPassAndSaltInput = String.Join(",",enteredPassword, storedSalt);
                        var joinedPassAndSaltStored = String.Join(",",storedHashPassword, storedSalt);
                        if (String.Equals(userName, storedUsername, StringComparison.OrdinalIgnoreCase))
                        {
                            var hashEnteredPassword = ComputeHash(joinedPassAndSaltInput);
                            var passWordCheck = String.Equals(hashEnteredPassword, storedHashPassword);
                            if (passWordCheck)
                            {
                                return true;
                            }
                            else
                            {
                                Console.WriteLine("Hash and Salt do not match");
                            }
                        }
                    }
                    else
                    {
                        Console.WriteLine("Hashed and Salt Password check failed");
                        return false;
                    }
                }
            }
            return false;
        }
        static string GetValidPassword()
        {
            string password = null;
            bool validPassword = false;

            while (!validPassword)
            {
                Console.WriteLine("Enter your password below. It must be all lowercase alphabetic characters with a minimum length of " +
                    $"{minPasswordLength} and a maximum length of {maxPasswordLength}.");
                Console.Write("Password: ");
                var firstPassword = GetPasswordInput();
                Console.WriteLine("Confirm Password: ");
                var finalPassword = GetPasswordInput();
                var passwordCheck = PasswordMatchConfirm(firstPassword, finalPassword);
                var isPasswordValid = IsLowerCaseAlpha(finalPassword, passwordFlag);

                if (passwordCheck && isPasswordValid)
                {
                    Console.WriteLine("Password Validated");
                    validPassword = true;
                    return password = finalPassword.Trim();
                    
                }
                else
                {
                    validPassword = false;
                    Console.WriteLine("Passwords are invalid or do not match. Please re-enter.");
                }
            }

            return password;
        }
        static string GenerateSalt()
        {
            byte[] salt = new byte[1];
            using(RNGCryptoServiceProvider rng = new RNGCryptoServiceProvider())
            {
                rng.GetBytes(salt);
            }

            return Convert.ToBase64String(salt);
        }
        static string ComputeHash(string input)
        {
            using(SHA256 sha256 = SHA256.Create())
            {
                using(SHA256 sHA256 = SHA256.Create())
                {
                    byte[] inputBytes = Encoding.UTF8.GetBytes(input);
                    byte[] hashBytes = sha256.ComputeHash(inputBytes);
                    StringBuilder sb = new StringBuilder();
                    for(int i = 0; i < hashBytes.Length; i++)
                    {
                        sb.Append(hashBytes[i].ToString("X2"));
                    }
                    
                    return sb.ToString();
                }
            }
        }
        static void WriteLog(string uName, string pWord,string fullPath)
        {
            using(StreamWriter sw = new StreamWriter(fullPath,true))
            {
                sw.WriteLine($"{uName},{pWord}");
            } 
        }
        static void WriteLogSalt(string uName, string hashAndSalt,string justSalt, string fullPath)
        {
            using (StreamWriter sw = new StreamWriter(fullPath, true))
            {
                sw.WriteLine($"{uName},{hashAndSalt},{justSalt}");
            }
        }
        static bool IsLowerCaseAlpha(string s,string passwordOrUserName)
        {
            var checkThis = s.Trim();
            switch (passwordOrUserName)
            {
                case ("username"):
                    if (string.IsNullOrEmpty(checkThis) || !Regex.IsMatch(checkThis, "^[a-z]+$") || checkThis.Length > 10)
                    {
                        return false;
                    }
                    break;
                case("password"):
                    if (string.IsNullOrEmpty(checkThis) || !Regex.IsMatch(checkThis, "^[a-z]+$") || 
                        checkThis.Length < minPasswordLength || checkThis.Length > maxPasswordLength)
                    {
                        return false;
                    }
                    break;
                default:
                    return false;
                    
            }
           return true;
            
        }
        static bool IsUsernameTaken(string username)
        {
            string[] fileNames = { plainTextUserCreds, hashedUserCreds, hashSaltUserCreds };

            foreach (string fileName in fileNames)
            {
                string filePath = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.UserProfile), "C:\\project2creds", fileName);

                try
                {
                   using (StreamReader reader = new StreamReader(filePath))
                   {
                     string line;
                     while ((line = reader.ReadLine()) != null)
                     {
                            string[] separatedString = line.Split(',');
                            if (separatedString.Length >= 1)
                            {
                                string extractUsername = separatedString[0].Trim();
                                if (extractUsername == username)
                                {
                                    return true;
                                }
                            }
                     }
                   }
                }catch (FileNotFoundException)
                 {
                    continue; // If the file doesn't exist, continue to the next file
                 }
                }

                return false; // Username is not taken in any of the files
            }
        static string GetPasswordInput()
        {
            StringBuilder sb = new StringBuilder();
            ConsoleKeyInfo key;
            do
            {
                key = Console.ReadKey(intercept: true);
                if (key.Key == ConsoleKey.Backspace && sb.Length > 0)
                {
                    sb.Remove(sb.Length - 1, 1);
                    Console.Write("\b \b");
                }
                else if (key.Key != ConsoleKey.Enter)
                {
                    sb.Append(key.KeyChar);
                    Console.Write("*");
                }
            } while (key.Key != ConsoleKey.Enter);
            Console.WriteLine();
            return sb.ToString();
        }
        static bool PasswordMatchConfirm(string firstEntry,string secondEntry)
        {
            string password = firstEntry.Trim();
            string confirmPassword = secondEntry.Trim();
            if (password != confirmPassword)
            {
                return false; //pw do not match
            }
            return true;
        }
        static void Main(string[] args)
        {
            
                Directory.CreateDirectory(userCdrivePath); //create file directory for later use
                string[] userFiles = { deliveryPathPlainTextFile, deliveryPathHashFile, deliveryPathHashSalt };
                foreach (string file in userFiles)
                {
                    if (!File.Exists(file))
                    {
                        Console.WriteLine("Creating User File " + file + "in " + userCdrivePath);
                        File.Create(file).Close();
                    }
                    else
                    {
                        continue;
                    }
                }

            var continueProgram = true;
            while (continueProgram == true)
            {
                Console.WriteLine("Do you want to create an account or login to an existing account?");
                Console.WriteLine("Press 1 to create and new account and 2 to login, or select 0 to exit ");
                int result = 0;
                bool inputValid = false;
                bool validIntParse = false;
                int verifyOrEnterPassword = 0;
                string userName = null;
                string password = null;
                bool accountsExist = false;

                //C:\Users\scott\source\repos\ConsoleApp4
                while (!validIntParse)
                {
                    string input = Console.ReadLine();
                    try
                    {
                        result = Int32.Parse(input);
                        validIntParse = true;
                    }
                    catch (FormatException)
                    {
                        Console.ForegroundColor = ConsoleColor.Red;
                        Console.BackgroundColor = ConsoleColor.Black;
                        Console.WriteLine("Invalid format, please try again.");
                        Console.ResetColor();
                        result = 0;
                        validIntParse = false;
                    }

                }
                foreach (string file in userFiles)
                {
                    if (File.Exists(file) && new FileInfo(file).Length > 0)
                    {
                        accountsExist = true;
                        break; // At least one account exists, no need to continue checking
                    }
                    else
                    {
                        accountsExist = false;
                    }
                }

                if (!accountsExist && result == 2)
                {
                    Console.ForegroundColor = ConsoleColor.Red;
                    Console.BackgroundColor = ConsoleColor.Black;
                    Console.WriteLine("No accounts exist. Select option 1 to create a new account.");
                    Console.ResetColor();
                    Environment.Exit(0);
                }else if(result == 0)
                {
                    Console.WriteLine("Exiting program");
                    continueProgram = false;
                    break;
                }
                else
                {
                    do
                    {

                        switch (result)
                        {
                            case 1:

                                Console.WriteLine("You selected to create a new account!");
                                Console.ForegroundColor = ConsoleColor.Red;
                                Console.BackgroundColor = ConsoleColor.Black;
                                Console.WriteLine("Please enter a username, only enter lowercase letters a-z, " +
                                                 "leading and trailing whitespaces will be removed.\n");
                                Console.ResetColor();
                                Console.Write("Username: ");
                                string enteredUsername = Console.ReadLine();
                                var isUserNameValid = IsLowerCaseAlpha(enteredUsername, userNameFlag);
                                var userNameTaken = IsUsernameTaken(enteredUsername);
                                if (isUserNameValid && (!userNameTaken))
                                {
                                    userName = enteredUsername.Trim();
                                    Console.WriteLine("\nUsername selected: " + userName);
                                    verifyOrEnterPassword = 0;
                                    inputValid = true;
                                    break;
                                }
                                else if (isUserNameValid && userNameTaken)
                                {
                                    Console.WriteLine("user name already taken, please enter another.");
                                    Console.ReadKey();
                                    inputValid = false;
                                    break;
                                }
                                else
                                {
                                    Console.WriteLine("Invalid user name, please limit to lower case a-z");
                                    inputValid = false;
                                    break;
                                }
                            case 2:
                                Console.WriteLine("Please enter your username: ");
                                var enteredUser = Console.ReadLine();
                                var userExists = IsUsernameTaken(enteredUser);
                                if (userExists)
                                {
                                    userName = enteredUser;
                                    verifyOrEnterPassword = 1;
                                    inputValid = true;
                                }
                                else if (!userExists)
                                {
                                    Console.ForegroundColor = ConsoleColor.Red;
                                    Console.BackgroundColor = ConsoleColor.Black;
                                    Console.WriteLine("Username does not exist. Please verify your username and re-enter it.");
                                    Console.ResetColor();
                                    inputValid = false;
                                }
                                break;
                            default:
                                inputValid = false;
                                Console.ForegroundColor = ConsoleColor.Red;
                                Console.BackgroundColor = ConsoleColor.Black;
                                Console.WriteLine("Invalid selection");
                                Console.ResetColor();
                                break;
                        }
                    } while (!inputValid);
                }
                if (verifyOrEnterPassword == 1 && accountsExist)
                {
                    bool passWordValid = false;
                    while (!passWordValid)
                    {
                        var finalPassword = GetValidPassword();
                        var isPasswordValid = IsLowerCaseAlpha(finalPassword, passwordFlag);
                        if (isPasswordValid)
                        {
                            Console.ForegroundColor = ConsoleColor.Green;
                            Console.BackgroundColor = ConsoleColor.Black;
                            Console.WriteLine("Password Enteries Match");
                            Console.ResetColor();
                            password = finalPassword.Trim();
                            passWordValid = true;
                        }
                        else
                        {
                            passWordValid = false;
                            Console.ForegroundColor = ConsoleColor.Red;
                            Console.BackgroundColor = ConsoleColor.Black;
                            Console.WriteLine("Passwords are invalid or do not match please re-enter: ");
                            Console.ResetColor();
                        }
                    }
                    
                }
                else if(result == 1 && verifyOrEnterPassword == 0)
                {
                    bool passWordValid = false;
                    while (!passWordValid)
                    {
                        var finalPassword = GetValidPassword();
                        var isPasswordValid = IsLowerCaseAlpha(finalPassword, passwordFlag);
                        if (isPasswordValid)
                        {
                            Console.ForegroundColor = ConsoleColor.Green;
                            Console.BackgroundColor = ConsoleColor.Black;
                            Console.WriteLine("Password Enteries Match");
                            Console.ResetColor();
                            password = finalPassword.Trim();
                            passWordValid = true;
                        }
                        else
                        {
                            passWordValid = false;
                            Console.ForegroundColor = ConsoleColor.Red;
                            Console.BackgroundColor = ConsoleColor.Black;
                            Console.WriteLine("Passwords are invalid or do not match please re-enter: ");
                            Console.ResetColor();
                        }
                    }
                    if (passWordValid)
                    {
                        Console.WriteLine("account created.");
                        WriteLog(userName, password, deliveryPathPlainTextFile);
                        string salt = GenerateSalt();
                        string hashAndSaltJoin = string.Join(",", password, salt);
                        string justHashedPw = ComputeHash(password);
                        string hashedPwWithSalt = ComputeHash(hashAndSaltJoin);
                        WriteLog(userName, justHashedPw, deliveryPathHashFile);
                        //WriteLog(userName, hashedPwWithSalt, deliveryPathHashSalt);//need to seperate hash and salt. then the password verification will be fixed.
                        WriteLogSalt(userName, hashedPwWithSalt, salt, deliveryPathHashSalt);
                        continueProgram = false;
                    }
                   
                }
                if(result ==2)
                {
                    //verify plaintext password logic here
                    var loginPassword = password;
                    if (ReadPassword(userName, loginPassword, deliveryPathPlainTextFile))
                    {
                        Console.WriteLine("Plain Text Login Sucessful! Press Enter to Continue");
                        Console.ReadKey();
                    }
                    else
                    {
                        Console.WriteLine("Plain Text Login Failed");
                    }

                    if (ReadPassword(userName, loginPassword, deliveryPathHashFile))
                    {
                        Console.WriteLine("Hash password Login Sucessful! Press Enter to Continue");
                        Console.ReadKey();
                    }
                    else
                    {
                        Console.WriteLine("Hashed Password Login Failed! Press Enter to Continue");
                        Console.ReadKey();
                    }

                    if (ReadPassword(userName, loginPassword, deliveryPathHashSalt))
                    {
                        Console.WriteLine("Hash & salt password Login Sucessful! Press Enter to Continue");
                        Console.ReadKey();
                        continueProgram = false;
                    }
                    else
                    {
                        Console.WriteLine("Hash & salt password Login Failed! Press Enter to Continue");
                        Console.ReadKey();
                        continueProgram = false;
                    }
                    
                }
            }
        }
    }
}

